import unittest

from ab import stats


class StatsTestCase(unittest.TestCase):

    def test_conversion_rate(self):
        test = (1000, 10)
        out = stats.conversion_rate(test)
        self.assertEqual(out, 0.01)

    def test_conversion_rate_2(self):
        test = (1000, 123)
        out = stats.conversion_rate(test)
        self.assertEqual(out, 0.123)

    def test_conversion_rate_uncertainty(self):
        test = (1000, 10)
        out = stats.conversion_rate_uncertainty(test)
        self.assertEqual(out, 0.003146426544510455)

    def test_conversion_rate_uncertainty_2(self):
        test = (1000, 123)
        out = stats.conversion_rate_uncertainty(test)
        self.assertEqual(out, 0.010386096475577337)

    def test_conversion_no_uplift(self):
        control = (1000, 10)
        test = (1000, 10)
        out = stats.conversion_uplift(control, test)
        self.assertEqual(out, 0)

    def test_conversion_uplift_test(self):
        control = (1000, 10)
        test = (1000, 50)
        out = stats.conversion_uplift(control, test)
        self.assertEqual(out, 4.0)

    def test_conversion_uplift_control(self):
        control = (1000, 50)
        test = (1000, 10)
        out = stats.conversion_uplift(control, test)
        self.assertEqual(out, -0.7999999999999999)

    def test_z_score_even(self):
        test = (1000, 10)
        test_rate = stats.conversion_rate(test)
        test_unc = stats.conversion_rate_uncertainty(test)
        control = (1000, 10)
        control_rate = stats.conversion_rate(control)
        control_unc = stats.conversion_rate_uncertainty(control)
        out = stats.z_score(test_rate, test_unc, control_rate, control_unc)
        self.assertEqual(out, 0.0)

    def test_z_score_variant_winning(self):
        test = (1000, 50)
        test_rate = stats.conversion_rate(test)
        test_unc = stats.conversion_rate_uncertainty(test)
        control = (1000, 10)
        control_rate = stats.conversion_rate(control)
        control_unc = stats.conversion_rate_uncertainty(control)
        out = stats.z_score(test_rate, test_unc, control_rate, control_unc)
        self.assertEqual(out, 5.279636773484547)

    def test_z_score_control_winning(self):
        test = (1000, 10)
        test_rate = stats.conversion_rate(test)
        test_unc = stats.conversion_rate_uncertainty(test)
        control = (1000, 50)
        control_rate = stats.conversion_rate(control)
        control_unc = stats.conversion_rate_uncertainty(control)
        out = stats.z_score(test_rate, test_unc, control_rate, control_unc)
        self.assertEqual(out, -5.279636773484547)

    def test_p_value_default(self):
        z_score = 0.0
        self.assertEqual(stats.p_value(z_score), 0.5)

    def test_p_value_variant_winning(self):
        test = (1000, 50)
        test_rate = stats.conversion_rate(test)
        test_unc = stats.conversion_rate_uncertainty(test)
        control = (1000, 10)
        control_rate = stats.conversion_rate(control)
        control_unc = stats.conversion_rate_uncertainty(control)
        z_score = stats.z_score(test_rate, test_unc, control_rate, control_unc)
        self.assertEqual(stats.p_value(z_score), 6.472011328743767e-08)

    def test_p_value_control_winning(self):
        test = (1000, 10)
        test_rate = stats.conversion_rate(test)
        test_unc = stats.conversion_rate_uncertainty(test)
        control = (1000, 50)
        control_rate = stats.conversion_rate(control)
        control_unc = stats.conversion_rate_uncertainty(control)
        z_score = stats.z_score(test_rate, test_unc, control_rate, control_unc)
        self.assertEqual(stats.p_value(z_score), 6.472011328743767e-08)

    def test_get_results_missing_control_key(self):
        with self.assertRaises(stats.StatsError):
            scores = {
                'variant1': (1000, 25),
                'variant2': (1000, 50),
            }
            stats.get_results(scores)

    def test_get_results(self):
        scores = {
            'control': (1000, 10),
            'variant1': (1000, 25),
            'variant2': (1000, 50),
        }
        data = stats.get_results(scores)
        self.assertEqual(len(data), 3)
        control = data['control']

        # Control
        self.assertEqual(len(control), 7)
        self.assertEqual(control.conversion_rate, 0.01)
        self.assertEqual(control.conversion_rate_uncertainty,
                         0.003146426544510455)
        self.assertEqual(control.uplift, 0.0)
        self.assertEqual(control.z_score, 0.0)
        self.assertEqual(control.p_value, 0.0)
        self.assertEqual(control.is_better_than_control, False)
        self.assertEqual(control.is_significant, False)

        # Variant1
        variant1 = data['variant1']
        self.assertEqual(len(variant1), 7)
        self.assertEqual(variant1.conversion_rate, 0.025)
        self.assertEqual(variant1.conversion_rate_uncertainty,
                         0.0049371044145328745)
        self.assertEqual(variant1.uplift, 1.5)
        self.assertEqual(variant1.z_score, 2.5621380568422687)
        self.assertEqual(variant1.p_value, 0.005201497829248816)
        self.assertEqual(variant1.is_better_than_control, True)
        self.assertEqual(variant1.is_significant, True)

        # Variant2
        variant2 = data['variant2']
        self.assertEqual(len(variant2), 7)
        self.assertEqual(variant2.conversion_rate, 0.05)
        self.assertEqual(variant2.conversion_rate_uncertainty,
                         0.006892024376045111)
        self.assertEqual(variant2.uplift, 4.0)
        self.assertEqual(variant2.z_score, 5.279636773484547)
        self.assertEqual(variant2.p_value, 6.472011328743767e-08)
        self.assertEqual(variant2.is_better_than_control, True)
        self.assertEqual(variant2.is_significant, True)



import collections
import uuid
import unittest

from ab import ab


class ABTestCase(unittest.TestCase):

    def test_calc_none_bucket(self):
        with self.assertRaisesRegex(ab.ABTestError, "saw <class 'NoneType'>"):
            ab.calc(None, buckets=2)

    def test_calc_negative_bucket(self):
        with self.assertRaisesRegex(ab.ABTestError, 'saw -1'):
            ab.calc(0, buckets=-1)

    def test_calc_0_bucket(self):
        with self.assertRaisesRegex(ab.ABTestError, 'saw 0'):
            ab.calc(0, buckets=0)

    def test_calc_1_bucket(self):
        with self.assertRaisesRegex(ab.ABTestError, 'saw 1'):
            ab.calc(0, buckets=1)

    def test_calc_zero_2_buckets(self):
        self.assertEqual(ab.calc(0, buckets=2), 1)

    def test_calc_2_buckets(self):
        self.assertEqual(ab.calc(99, buckets=2), 1)
        self.assertEqual(ab.calc(98, buckets=2), 0)

    def test_calc_3_buckets(self):
        self.assertEqual(ab.calc(99, buckets=3), 1)
        self.assertEqual(ab.calc(98, buckets=3), 0)

    def test_calc_100_buckets(self):
        self.assertEqual(ab.calc(99, buckets=100), 60)

    def test_100_users(self):
        user_ids = list(range(100))
        allocations = ab.allocate(user_ids, buckets=2)
        allocation_keys = sorted([int(k) for k in allocations.keys()])
        self.assertEqual(allocation_keys, list(range(0, 2)))
        self.assertEqual(allocations[0], 43)
        self.assertEqual(allocations[1], 57)

    def test_1000_users(self):
        user_ids = list(range(1000))
        allocations = ab.allocate(user_ids, buckets=2)
        allocation_keys = sorted([int(k) for k in allocations.keys()])
        self.assertEqual(allocation_keys, list(range(0, 2)))
        self.assertEqual(allocations[0], 481)
        self.assertEqual(allocations[1], 519)

    def test_one_million_emails(self):
        emails = [str(uuid.uuid4()) for i in range(1000000)]
        allocations = ab.allocate(emails, buckets=100)
        allocation_keys = sorted([int(k) for k in allocations.keys()])
        self.assertEqual(allocation_keys, list(range(0, 100)))
        for value in allocations.values():
            int_value = int(value)
            self.assertGreaterEqual(int_value, 9500)
            self.assertLessEqual(int_value, 10500)

    def test_get_bucket_example(self):
        user_id = 1
        test_name = 'my_test_v1'
        buckets = ['control', 'variant1', 'variant2']
        allocation = ab.get_bucket(user_id, test=test_name, buckets=buckets)
        self.assertEqual(allocation, 'control')

    def test_get_bucket_test_name_changes_allocation(self):
        user_id = 1
        test_name = 'my_test_v2'  # difference
        buckets = ['control', 'variant1', 'variant2']
        allocation = ab.get_bucket(user_id, test=test_name, buckets=buckets)
        self.assertEqual(allocation, 'variant1')

    def test_get_bucket_example_1000_users(self):
        allocations = collections.defaultdict(int)
        test_name = 'my_test_v2'
        buckets = ['control', 'variant1', 'variant2']
        for user_id in range(1000):
            bucket = ab.get_bucket(user_id, test=test_name, buckets=buckets)
            allocations[bucket] += 1
        self.assertEqual(allocations['control'], 356)
        self.assertEqual(allocations['variant1'], 310)
        self.assertEqual(allocations['variant2'], 334)



class MockRedis(object):

    def __init__(self):
        self.data = {}

    def set(self, key, val):
        self.data[key] = val
        return val

    def get(self, key):
        return self.data.get(key)

    def incr(self, key):
        val = int(self.get(key) or '0')
        return self.set(key, val + 1)

    def mget(self, keys):
        return [self.get(key) for key in keys]



from typing import Dict, Tuple, NamedTuple
import math
import scipy.stats

HIGH_CONFIDENCE = 0.05
CONTROL_KEY = 'control'


# PUBLIC


class Data(NamedTuple):
    conversion_rate: float
    conversion_rate_uncertainty: float
    uplift: float
    z_score: float
    p_value: float
    is_better_than_control: bool
    is_significant: bool
    # TODO(DAN): is_winner: bool


def get_results(
        scores: Dict[str, Tuple[int, int]],
        control_key: str = CONTROL_KEY,
        p: float = HIGH_CONFIDENCE):
    assert_control_key(scores)
    control = scores[CONTROL_KEY]
    cr_c = conversion_rate(control)
    cru_c = conversion_rate_uncertainty(control)

    data = {}
    data[CONTROL_KEY] = Data(
        conversion_rate=cr_c,  # TODO(DAN): ± format
        conversion_rate_uncertainty=cru_c,
        uplift=0.0,
        z_score=0.0,
        p_value=0.0,
        is_better_than_control=False,
        is_significant=False)

    for bucket, test in scores.items():
        if bucket == CONTROL_KEY:
            continue
        cr_t = conversion_rate(test)
        cru_t = conversion_rate_uncertainty(test)
        uplift = conversion_uplift(control, test)
        z_t = z_score(cr_t, cru_t, cr_c, cru_c)
        p_t = p_value(z_t)
        is_better_than_control = uplift > 0
        is_significant = p_t < p
        data[bucket] = Data(
            conversion_rate=cr_t,  # TODO(DAN): ± format
            conversion_rate_uncertainty=cru_t,
            uplift=uplift,
            z_score=z_t,
            p_value=p_t,
            is_better_than_control=is_better_than_control,
            is_significant=is_significant)

    return data

# PRIVATE


class StatsError(Exception):
    pass


def conversion_rate(test: Tuple[float]) -> float:
    """Successes versus Trials."""
    return test[1] / test[0]


def conversion_rate_uncertainty(test: Tuple[float]) -> float:
    """Standard error rate."""
    rate = conversion_rate(test)
    return math.sqrt((rate * (1 - rate)) / test[0])


def conversion_uplift(control: Tuple[float],
                      test: Tuple[float]) -> float:
    """Determine if test is better than control."""
    control_rate = conversion_rate(control)
    test_rate = conversion_rate(test)
    return (test_rate - control_rate) / control_rate


def z_score(test_conversion_rate: float,
            test_uncertainty_rate: float,
            control_conversion_rate: float,
            control_uncertainty_rate: float) -> float:
    """Standard deviations away from mean value."""
    numerator = test_conversion_rate - control_conversion_rate
    denominator = math.sqrt(
        test_uncertainty_rate ** 2 + control_uncertainty_rate ** 2)
    return numerator / denominator


def p_value(z_score: float) -> float:
    """Statistical significance. Lower = better."""
    return scipy.stats.norm.sf(abs(z_score))


def assert_control_key(scores: Dict[str, Tuple[int, int]]):
    if CONTROL_KEY not in scores:
        raise StatsError(f'{CONTROL_KEY} is required to test results')



from typing import List, Tuple, Dict
import logging
import random

from ab import store

NAMESPACE = 'mab'
KEY_TRIALS = 'trials'
KEY_SUCCESSES = 'successes'

# 10% explore, 90% exploit
RATE = .1


# PUBLIC

def get_bucket(test: str, buckets: List[str], rate: float = RATE) -> str:
    """Obtain a bucket for a test with multi-arm bandit strategy."""
    assert_explore_rate(rate)
    assert_buckets(buckets)
    chance = random.random()  # 0.0-1.0
    if chance <= rate:
        return explore(test, buckets)
    return exploit(test, buckets)


def success(test: str, bucket: str):
    """Mark a trial success for a test."""
    try:
        return store.incr(format_key(test, bucket, KEY_SUCCESSES))
    except store.StoreError as e:
        logging.exception(e)
        return None


class MABTestError(Exception):
    pass


# PRIVATE


def assert_explore_rate(rate: float):
    if 0.0 < rate > 1.0:
        message = f'Explore rate must be between 0.0 and 1.0 (saw: {rate})'
        raise MABTestError(message)


def assert_buckets(buckets: List[str]):
    if len(buckets) < 2:
        raise MABTestError(f'Must have two or more buckets (saw {buckets})')


def format_key(test: str, bucket: str, key: str) -> str:
    return f'{NAMESPACE}:{test}:{bucket}:{key}'


def trial(test: str, bucket: str):
    return store.incr(format_key(test, bucket, KEY_TRIALS))


def get_state(test: str, buckets: List[str]) -> Dict[str, int]:
    keys = []
    for bucket in buckets:
        keys.append(format_key(test, bucket, KEY_TRIALS))
        keys.append(format_key(test, bucket, KEY_SUCCESSES))
    payload = store.mget(keys)
    data, index = {}, 0
    for bucket in buckets:
        data[bucket] = {
            KEY_TRIALS: int(payload[index] or 0),
            KEY_SUCCESSES: int(payload[index + 1] or 0)}
        index += 2
    return data


def get_state_with_tuples(
        test: str, buckets: List[str]) -> Dict[str, Tuple[int, int]]:
    return {bucket: (item[KEY_TRIALS], item[KEY_SUCCESSES])
            for bucket, item in get_state(test, buckets).items()}


def get_scores(test: str, buckets: List[str]) -> Dict[str, int]:
    return {key: stats[KEY_SUCCESSES] / (stats[KEY_TRIALS] or 1)
            for key, stats in get_state(test, buckets).items()}


def explore(test: str, buckets: List[str]) -> str:
    bucket = random.choice(buckets)
    trial(test, bucket)
    return bucket


def exploit(test: str, buckets: List[str]) -> str:
    # TODO(DAN): use stats.get_results() to pick optimal bucket.
    scores = get_scores(test, buckets)

    # If all buckets are even, explore.
    if len(set(scores.values())) == 1:
        return explore(test, buckets)

    # Exploit leading bucket.
    bucket = max(scores, key=scores.get)
    trial(test, bucket)
    return bucket



from ab import mab

TEST_NAME = 'MY_DEMO_TEST_V1'
TEST_CONTROL = 'control'
TEST_VARIANT1 = 'variant1'
TEST_VARIANT2 = 'variant2'
TEST_BUCKETS = (
    TEST_CONTROL,
    TEST_VARIANT1,
    TEST_VARIANT2,
)

while True:
    bucket = mab.get_bucket(TEST_NAME, TEST_BUCKETS)
    try:
        val = input(f'Did {bucket} succeed? y/n: ').strip()
    except KeyboardInterrupt:
        exit()
    if val == 'y':
        mab.success(TEST_NAME, bucket)
    print(mab.get_scores(TEST_NAME, TEST_BUCKETS))



from typing import List, Dict
import collections
import hashlib

HASH_LEN = 16
MAX_HASH = HASH_LEN**40
HASH_ENCODING = 'utf-8'


# PUBLIC

def get_bucket(value, test: str, buckets: List[str]) -> str:
    """Given a value for a unique test name, determine the bucket.
    Test name is factored into allocation to ensure allocations
    are unique. Without this, user_id 1 could always end up in the
    same bucket, every time you ran a new experiment.
    """
    buckets_len = len(buckets)
    test_name_hash = hash_value(test)
    input_value = f'{value}{test_name_hash}'
    allocation = calc(input_value, buckets_len)
    return buckets[allocation]


class ABTestError(Exception):
    pass


# PRIVATE


def assert_buckets(buckets: int):
    if buckets < 2:
        raise ABTestError(f'Must have two or more buckets (saw {buckets})')


def assert_value(value):
    if not isinstance(value, (int, str)):
        value_type = type(value)
        raise ABTestError(f'Value must be a string or int (saw {value_type})')


def hash_value(value) -> int:
    assert_value(value)
    safe_value = str(value).encode(HASH_ENCODING)
    return int(hashlib.sha1(safe_value).hexdigest(), HASH_LEN)


def calc(value, buckets: int) -> int:
    """Given a value, calculate the bucket index."""
    assert_buckets(buckets)
    val_hash = hash_value(value)
    div = MAX_HASH / buckets
    return int(float(val_hash / div))


def allocate(values, buckets: int) -> Dict[str, str]:
    """Allocate values into named buckets by index.
    This is useful for verifying that distribution is even.
    """
    allocations = collections.defaultdict(int)
    for i in values:
        bucket = calc(i, buckets)
        allocations[bucket] += 1
    return allocations



import random
import collections
import unittest
from unittest import mock

from ab import store_mock
from ab import mab
from ab import stats


class MultiArmedBanditTestCase(unittest.TestCase):

    def test_trial(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            self.assertEqual(mab.trial('my_test_v1', 'control'), 1)
            self.assertEqual(mab.trial('my_test_v1', 'control'), 2)

    def test_success(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            self.assertEqual(mab.success('my_test_v1', 'control'), 1)
            self.assertEqual(mab.success('my_test_v1', 'control'), 2)

    def test_get_state_default(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            test_name = 'my_test_v1'
            buckets = ['control', 'variant1', 'variant2']
            data = mab.get_state(test_name, buckets)
            self.assertEqual(list(data.keys()), buckets)
            self.assertEqual(data['control'][mab.KEY_TRIALS], 0)
            self.assertEqual(data['control'][mab.KEY_SUCCESSES], 0)
            self.assertEqual(data['variant1'][mab.KEY_TRIALS], 0)
            self.assertEqual(data['variant1'][mab.KEY_SUCCESSES], 0)
            self.assertEqual(data['variant2'][mab.KEY_TRIALS], 0)
            self.assertEqual(data['variant2'][mab.KEY_SUCCESSES], 0)

    def test_get_state_with_tuples_default(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            test_name = 'my_test_v1'
            buckets = ['control', 'variant1', 'variant2']
            data = mab.get_state_with_tuples(test_name, buckets)
            self.assertEqual(list(data.keys()), buckets)
            self.assertEqual(data['control'][0], 0)
            self.assertEqual(data['control'][1], 0)
            self.assertEqual(data['variant1'][0], 0)
            self.assertEqual(data['variant1'][1], 0)
            self.assertEqual(data['variant2'][0], 0)
            self.assertEqual(data['variant2'][1], 0)

    def test_get_state_used(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            test_name = 'my_test_v1'
            buckets = ['control', 'variant1', 'variant2']
            mab.trial('my_test_v1', 'control')
            mab.trial('my_test_v1', 'control')
            mab.success('my_test_v1', 'control')
            mab.success('my_test_v1', 'control')
            mab.trial('my_test_v1', 'variant1')
            mab.success('my_test_v1', 'variant1')
            mab.trial('my_test_v1', 'variant2')
            data = mab.get_state(test_name, buckets)
            self.assertEqual(list(data.keys()), buckets)
            self.assertEqual(data['control'][mab.KEY_TRIALS], 2)
            self.assertEqual(data['control'][mab.KEY_SUCCESSES], 2)
            self.assertEqual(data['variant1'][mab.KEY_TRIALS], 1)
            self.assertEqual(data['variant1'][mab.KEY_SUCCESSES], 1)
            self.assertEqual(data['variant2'][mab.KEY_TRIALS], 1)
            self.assertEqual(data['variant2'][mab.KEY_SUCCESSES], 0)

    def test_get_state_with_tuples_used(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            test_name = 'my_test_v1'
            buckets = ['control', 'variant1', 'variant2']
            mab.trial('my_test_v1', 'control')
            mab.trial('my_test_v1', 'control')
            mab.success('my_test_v1', 'control')
            mab.success('my_test_v1', 'control')
            mab.trial('my_test_v1', 'variant1')
            mab.success('my_test_v1', 'variant1')
            mab.trial('my_test_v1', 'variant2')
            data = mab.get_state_with_tuples(test_name, buckets)
            self.assertEqual(list(data.keys()), buckets)
            self.assertEqual(data['control'][0], 2)
            self.assertEqual(data['control'][1], 2)
            self.assertEqual(data['variant1'][0], 1)
            self.assertEqual(data['variant1'][1], 1)
            self.assertEqual(data['variant2'][0], 1)
            self.assertEqual(data['variant2'][1], 0)

    def test_cold_start_exploit_explores(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            test_name = 'my_test_v1'
            buckets = ['control', 'variant1', 'variant2']
            seen = collections.Counter()
            for x in range(1000):
                bucket = mab.get_bucket(test_name, buckets)
                seen[bucket] += 1
                self.assertTrue(bucket in buckets)
            self.assertGreaterEqual(seen['control'], 275)
            self.assertGreaterEqual(seen['variant1'], 275)
            self.assertGreaterEqual(seen['variant2'], 275)
            self.assertEqual(sum(seen.values()), 1000)

    def test_exploit_variant_1_single_trial_still_exploring(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            test_name = 'my_test_v1'
            buckets = ['control', 'variant1', 'variant2']
            seen = collections.Counter()
            mab.trial(test_name, 'variant1')  # first trial with no success
            for x in range(1000):
                bucket = mab.get_bucket(test_name, buckets)
                seen[bucket] += 1
                self.assertTrue(bucket in buckets)
            self.assertGreaterEqual(seen['control'], 275)
            self.assertGreaterEqual(seen['variant1'], 275)
            self.assertGreaterEqual(seen['variant2'], 275)
            self.assertEqual(sum(seen.values()), 1000)

    def test_exploit_variant_1_three_trials_still_exploring(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            test_name = 'my_test_v1'
            buckets = ['control', 'variant1', 'variant2']
            seen = collections.Counter()
            mab.trial(test_name, 'variant1')  # first trial with no success
            mab.trial(test_name, 'variant1')  # second trial with no success
            mab.trial(test_name, 'variant1')  # third trial with no success
            for x in range(1000):
                bucket = mab.get_bucket(test_name, buckets)
                seen[bucket] += 1
                self.assertTrue(bucket in buckets)
            self.assertGreaterEqual(seen['control'], 250)
            self.assertGreaterEqual(seen['variant1'], 250)
            self.assertGreaterEqual(seen['variant2'], 250)
            self.assertEqual(sum(seen.values()), 1000)

    def test_exploit_variant_1_one_trial_with_begins_exploitation(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            test_name = 'my_test_v1'
            buckets = ['control', 'variant1', 'variant2']
            seen = collections.Counter()
            mab.trial(test_name, 'variant1')
            mab.success(test_name, 'variant1')  # first trial with success
            for x in range(1000):
                bucket = mab.get_bucket(test_name, buckets)
                seen[bucket] += 1
                self.assertTrue(bucket in buckets)
            self.assertGreaterEqual(seen['control'], 10)
            self.assertGreaterEqual(seen['variant1'], 850)
            self.assertGreaterEqual(seen['variant2'], 10)
            self.assertEqual(sum(seen.values()), 1000)

    def test_exploit_two_variants_with_success(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            test_name = 'my_test_v1'
            buckets = ['control', 'variant1', 'variant2']
            seen = collections.Counter()
            mab.trial(test_name, 'variant1')
            mab.success(test_name, 'variant1')  # first trial with success
            mab.trial(test_name, 'variant2')
            mab.success(test_name, 'variant2')  # other trial with success
            for x in range(1000):
                bucket = mab.get_bucket(test_name, buckets)
                seen[bucket] += 1
                self.assertTrue(bucket in buckets)
            self.assertGreaterEqual(seen['control'], 10)
            self.assertGreaterEqual(seen['variant1'], 400)
            self.assertGreaterEqual(seen['variant2'], 400)
            self.assertEqual(sum(seen.values()), 1000)

    def test_exploit_all_buckets_with_success(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            test_name = 'my_test_v1'
            buckets = ['control', 'variant1', 'variant2']
            seen = collections.Counter()
            mab.trial(test_name, 'control')
            mab.success(test_name, 'control')  # first trial with success
            mab.trial(test_name, 'variant1')
            mab.success(test_name, 'variant1')  # second trial with success
            mab.trial(test_name, 'variant2')
            mab.success(test_name, 'variant2')  # third trial with success
            for x in range(1000):
                bucket = mab.get_bucket(test_name, buckets)
                seen[bucket] += 1
                self.assertTrue(bucket in buckets)
            self.assertGreaterEqual(seen['control'], 250)
            self.assertGreaterEqual(seen['variant1'], 250)
            self.assertGreaterEqual(seen['variant2'], 250)
            self.assertEqual(sum(seen.values()), 1000)

    def test_variant_dominates(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            test_name = 'my_test_v1'
            buckets = ['control', 'variant1', 'variant2']
            seen = collections.Counter()
            for x in range(3):  # poor showing, control...
                mab.trial(test_name, 'control')
                mab.success(test_name, 'control')
            for x in range(9):  # Not bad, variant1...
                mab.trial(test_name, 'variant2')
                mab.success(test_name, 'variant2')
            for x in range(100):  # Variant2 leading the way.
                mab.trial(test_name, 'variant2')
                mab.success(test_name, 'variant2')
            for x in range(1000):
                bucket = mab.get_bucket(test_name, buckets)
                seen[bucket] += 1
                self.assertTrue(bucket in buckets)
            self.assertGreaterEqual(seen['control'], 15)
            self.assertLessEqual(seen['control'], 75)
            self.assertGreaterEqual(seen['variant1'], 15)
            self.assertLessEqual(seen['variant1'], 75)
            self.assertGreaterEqual(seen['variant2'], 850)
            self.assertLessEqual(seen['variant2'], 990)
            self.assertEqual(sum(seen.values()), 1000)

    def test_variant_dominates_low_exploration(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            test_name = 'my_test_v1'
            buckets = ['control', 'variant1', 'variant2']
            seen = collections.Counter()
            for x in range(3):  # poor showing, control...
                mab.trial(test_name, 'control')
                mab.success(test_name, 'control')
            for x in range(9):  # Not bad, variant1...
                mab.trial(test_name, 'variant2')
                mab.success(test_name, 'variant2')
            for x in range(100):  # Variant2 leading the way.
                mab.trial(test_name, 'variant2')
                mab.success(test_name, 'variant2')
            for x in range(1000):
                bucket = mab.get_bucket(test_name, buckets, rate=.05)
                seen[bucket] += 1
                self.assertTrue(bucket in buckets)
            self.assertGreaterEqual(seen['control'], 5)
            self.assertLessEqual(seen['control'], 75)
            self.assertGreaterEqual(seen['variant1'], 5)
            self.assertLessEqual(seen['variant1'], 75)
            self.assertGreaterEqual(seen['variant2'], 900)
            self.assertLessEqual(seen['variant2'], 990)
            self.assertEqual(sum(seen.values()), 1000)

    # I stress tested this and it always seems to pass.
    # If it breaks, enable this because random is random.
    # @unittest.skip('comment & run: make stress_test_mab')
    def test_simulate_experiment(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            test_name = 'my_test_v1'
            buckets = ['control', 'variant1', 'variant2']
            # Trying to simulate random 30% trial, random 10% success rate.
            for x in range(10000):
                # ~30% in trial
                if random.random() <= .3:
                    bucket = mab.get_bucket(test_name, buckets)
                    # ~10% of trial members succeed.
                    if random.random() <= .1:
                        mab.success(test_name, bucket)
            scores = mab.get_scores(test_name, buckets)
            # Leaving headroom for random
            self.assertLessEqual(scores['control'], .15)
            self.assertGreaterEqual(scores['control'], .01)
            self.assertLessEqual(scores['variant1'], .15)
            self.assertGreaterEqual(scores['variant1'], .01)
            self.assertLessEqual(scores['variant2'], .15)
            self.assertGreaterEqual(scores['variant2'], .01)
            state = mab.get_state_with_tuples(test_name, buckets)
            results = stats.get_results(state)
            self.assertGreaterEqual(results['variant1'].p_value, 0.0)
            self.assertLessEqual(results['variant1'].p_value, 0.5)

    # I stress tested this and it always seems to pass.
    # If it breaks, enable this because random is random.
    # @unittest.skip('comment & run: make stress_test_mab')
    def test_simulate_experiment_variant1_outperforms(self):
        with mock.patch('ab.store.get') as store_patch:
            store_patch.return_value = store_mock.MockRedis()
            test_name = 'my_test_v1'
            buckets = ['control', 'variant1', 'variant2']
            # Trying to simulate random 30% trial, random 10% success rate.
            for x in range(10000):
                if random.random() <= .3:  # ~30% in trial
                    bucket = mab.get_bucket(test_name, buckets)
                    # Variant 1 outperforming at ~30% success rate
                    success_chance = random.random()
                    if bucket == 'variant1' and success_chance <= .3:
                        mab.success(test_name, bucket)
                    # Only 5% of control and variant 2 members succeed.
                    elif success_chance <= .05:
                        mab.success(test_name, bucket)
            scores = mab.get_scores(test_name, buckets)
            self.assertGreaterEqual(scores['control'], .0)
            self.assertLessEqual(scores['control'], .15)
            # Variant 1 consistently does better.
            self.assertGreaterEqual(scores['variant1'], .15)
            self.assertLessEqual(scores['variant1'], .35)
            self.assertGreaterEqual(scores['variant2'], .0)
            self.assertLessEqual(scores['variant2'], .15)
            state = mab.get_state_with_tuples(test_name, buckets)
            results = stats.get_results(state)
            self.assertTrue(results['variant1'].is_better_than_control)
            self.assertTrue(results['variant1'].is_significant)
            self.assertGreaterEqual(results['variant1'].p_value, 0.0)
            self.assertLessEqual(results['variant1'].p_value, 0.001)



from distutils.core import setup

setup(
    name='ab',
    author='Dan Masquelier',
    author_email='dancrew32+ab@gmail.com',
    url='https://github.com/dancrew32/ab/',
    version='0.1dev',
    packages=['ab'],
    license='Unlicense',
    description='Easy A/B Test Bucketing',
)



from typing import List
import os

import redis

HOST = os.environ.get('AB_HOST', 'localhost')
PORT = int(os.environ.get('AB_PORT', 6379))
DB = int(os.environ.get('AB_DB', 0))

CONN = None


class StoreError(redis.exceptions.RedisError):
    pass


def get():
    global CONN
    if not CONN:
        CONN = redis.Redis(host=HOST, port=PORT, db=DB)
    return CONN


def incr(key: str):
    try:
        return get().incr(key)
    except redis.exceptions.RedisError as e:
        raise StoreError(e, advise())


def mget(keys: List[str]):
    try:
        return get().mget(keys)
    except redis.exceptions.RedisError as e:
        raise StoreError(e, advise())


def advise():
    return 'Please configure & run redis: make redis_install redis_start'
